package cscie160.project.atm;

import java.rmi.*;
import java.rmi.server.UnicastRemoteObject;
import java.util.*;

/** Project libraries */
import cscie160.project.*;
import cscie160.project.atm.*;
import cscie160.project.account.*;
import cscie160.project.bank.*;
import cscie160.project.security.*;
import cscie160.project.security.exception.*;



/**
  * By David Cheung <p>
  * CSCIE160 Project <p>
  * 05-16-2001 <p>
  *
  * ATMImpl implements ATM. At init time it will
  * bind to Bank and Security interface. Also, it will
  * create a vector to store callback objects.
  * <p>
  * To do (from specs in project):
  * 		1)notify all registered listeners of the 
  * 		transaction.<p>
  *		2) use security service to authenticate account info.<p>
  *		3) use security service to authorize operation on account.<p>
  *		4) Use bank to obtain account references<p>
  *		5) Use account references to perform transaction<p>
  * 
  */

public class ATMImpl
			extends UnicastRemoteObject
			implements ATM
			
{
	// objects
	/** bank remote object */
	private Bank BankRemote;
	/** Security remote object */
	private Security SecRemote;
	/** Amount of cashAtHand in the ATM */
	private float cashAtHand;
	// callback objects:
	private Vector callbackObjects;
	
	/**
	  * Adds call back.
	  */
	public void addCallBack(ATMListener aListener)
		throws RemoteException
	{
		callbackObjects.addElement(aListener);
		
	}
	
	/**
	  * Sends notification method:
	  * when this is called, it will take in a TransactionNotification
	  * object and send it to everyone in the list.
	  * It will pick up the ATMListener client and invoke the
	  * getNotify method. The getNotify method basically prints out
	  * the message.
	  */
	public void notifyAll(String message)
	{
		TransactionNotification noteBox = new TransactionNotification(message);
		for(int x=0; x< callbackObjects.size(); x++)
		{
			// convert the vector object to ATMListener object
			ATMListener callback = (ATMListener)callbackObjects.elementAt(x);
			try
			{
			// invoke the callback:
			callback.getNotify(noteBox);
			}
			catch(RemoteException re)
			{
				System.out.println("callback object not there...?");
			}
		}
	}
   /**
     * Constructor:
     * creates a linkedlist and 3 accounts.
     */
	public ATMImpl() 
			throws RemoteException
	{
		super();
		/** establish initial amount of money in ATM */
		cashAtHand=500;
		/** initialize callbeckObjects vector */
		callbackObjects = new Vector();
		
		/** lookup bank and security object */
		try
		{
				
			BankRemote = (Bank)Naming.lookup("//localhost/Bank");
			SecRemote = (Security)Naming.lookup("//localhost/Security");
		}
		catch(Exception e)
		{
			System.out.println("Some exception! ");
			e.printStackTrace();
		}

	}
	
	/**
	  * Deposit <p>
	  * Deposits money unto the account. <p>
	  * List of things to do:
	  * 		1)notify all registered listeners of the 
	  * 		transaction. <p>
     *		2) use security service to authenticate account info <p>
	  *		3) use security service to authorize operation on account. <p>
	  *		4) Use bank to obtain account references. <p>
	  *		5) Use account references to perform transaction <p>
	  *		6) catch throws various exceptions when neccessary.
	  * <p>
	  * @param		userNum: account number
	  * @param 		amount: amount to deposit
	  */
	public void deposit(AccountInfo anAccount, float amount) 
			throws java.rmi.RemoteException,
					UnpermittedTransactionException,
					SecurityInvalidPinException
	{
		/** authenticate account: */
		if(!SecRemote.verifyAccount(anAccount))
		{	
			System.out.println("Account has invalid PIN! ...throwing exception.");
			throw new SecurityInvalidPinException("Invalid Account PIN Exception thrown");
			
		}
		/** verifiy that the transaction is allowed. throw exception otherwise*/
		if(!SecRemote.getDepPermission(anAccount))
		{
			/** message appears in the ATM server. */
			System.out.println("Account does not have permission to deposit! ... throwing exception.");
			throw new UnpermittedTransactionException("Deposit Permission denied! exception thrown");
		}
		/** at this point everything should be ok...now try to get the account. */
		Account accountA;
		/** synchronize this for multiple user capability. */
		synchronized(this)
		{
			accountA = BankRemote.getAccount(anAccount.getAccountNum());
			accountA.deposit(amount);
			notifyAll("Account "+anAccount.getAccountNum()+ " has deposited $"+
							amount);

			/** increase amount of money stored in the ATM. */
			cashAtHand += amount;
		}
	}
	/**
	  * withdraw <p>
	  * Withdraws money from a particular account <p>
	  * <p>
	  * This will do: <p>
	  * 		1)notify all registered listeners of the 
	  * 		transaction. <p>
	  *		2) use security service to authenticate account info <p>
	  *		3) use security service to authorize operation on account. <p>
	  *		4) Use bank to obtain account references <p>
	  *		5) Use account references to perform transaction <p>
	  *		6) Throws exceptioni if neccessary
	  * 
	  * @param 		AccountInfo: account number and a PIn.
	  * @param		amount: amount to withdraw.
	  */
	public void withdraw(AccountInfo anAccount, float amount) 
			throws java.rmi.RemoteException,
					UnpermittedTransactionException,
					OverdrawnException,
					SecurityInvalidPinException
	{
		/** authenticate account. Throw exception otherwise. */
		if(!SecRemote.verifyAccount(anAccount))
		{	
			System.out.println("Account has invalid PIN! ...throwing exception.");
			throw new SecurityInvalidPinException("Invalid Account PIN Exception thrown");
			
			
		}
		
		/** verifiy transaction allowed. Throws exception otherwise.*/
		if(!SecRemote.getWithPermission(anAccount))
		{

			System.out.println("Account does not have permission to withdraw! ... throwing exception.");
			throw new UnpermittedTransactionException("Withdraw Permission denied! exception thrown");
		}
		// everything seems ok...now try to get the account.
		Account accountA;
		accountA = BankRemote.getAccount(anAccount.getAccountNum());
		// throw exception if balance is less than money withdrawn.
		System.out.print("Withdrawing $"+amount + " from balance $" +accountA.getBalance() +"\n");
		if( amount > accountA.getBalance())
		{
			//Debug:
			// This is displayed on server end.
			System.out.println("You are attempting to withdraw more than you have! ...throwning exception.\n");
			throw new OverdrawnException("You cannot withdraw more than you have! Exception thrown!\n");
		}
		if ( amount > cashAtHand)
		{
			System.out.println("ATM does not have enough money! throwning exception.\n");
			throw new OverdrawnException("ATM Does not have enough money! Exception thrown.\n");
		}
		/** synchronize this for multiple user capability. */		
		synchronized(this)
		{
			accountA.withdraw(amount);
		
			/** notify!! */
			notifyAll("Account " + anAccount.getAccountNum()+ " has withdrawn $"
					+ amount);
			/** adjust amount of money in ATM */
			cashAtHand-= amount;
		}
	}
	/**
	  * getBalance <p>
	  * Returns the balance of the account.<p>
	  * <p>
	  * @param		userNum: AccountInfo
	  */
	public float getBalance(AccountInfo anAccount) 
			throws java.rmi.RemoteException,
					UnpermittedTransactionException,
					SecurityInvalidPinException
	{
		/** Verify account pin is correct. Throw exception otherwise. */
		if(!SecRemote.verifyAccount(anAccount))
		{	
			System.out.println("Account has invalid PIN! ...throwing exception.");
			throw new SecurityInvalidPinException("Invalid Account PIN Exception thrown");
			
		}
		/** Verifiy transaction is allowed. Throws exception otherwise. */
		if(!SecRemote.getBalPermission(anAccount))
		{
		
			System.out.println("Account does not have permission to get balance! ... throwing exception.");
			throw new UnpermittedTransactionException("Get balance Permission denied! exception thrown");
		}
		/** Everything seems ok...now try to get the account. */
		Account accountA;
		accountA = BankRemote.getAccount(anAccount.getAccountNum());
		return accountA.getBalance();
	}
	
	/**
	  * transfer <p>
	  * Withdraws money from the fromAccount and deposits it to the toAccount.<p>
	  * <p>
	  * This will do: <p>
	  * 		1)notify all registered listeners of the 
	  * 		transaction. <p>
	  *		2) use security service to authenticate account info <p>
	  *		3) use security service to authorize operation on account. <p>
	  *		4) Use bank to obtain account references <p>
	  *		5) Use account references to perform transaction <p>
	  *		6) Throws exceptioni if neccessary
	  * 
	  * @param 		AccountInfo: account number and a Pin of from account.
	  * @param 		AccountInfo: account number and pin of to account.
	  * @param		amount: amount to transfer.
	  */

	public void transfer(AccountInfo fromAccount, AccountInfo toAccount, float amount) 
			throws java.rmi.RemoteException,
					UnpermittedTransactionException,
					OverdrawnException,
					SecurityInvalidPinException
	{
		
		/** authenticate account. Throws exception otherwise */
		if(!(SecRemote.verifyAccount(fromAccount)&& SecRemote.verifyAccount(toAccount)))
		{	
			
			System.out.println("Account has invalid PIN! ...throwing exception.");
			throw new SecurityInvalidPinException("Invalid Account PIN Exception thrown");
			
		}
		/** verifiy transaction allowed. Throws exception otherwise. */
		if(!SecRemote.getWithPermission(fromAccount))
		{
			System.out.println("Account does not have permission to withdraw! ... throwing exception.");
			throw new UnpermittedTransactionException("Withdraw Permission denied! exception thrown");
		}
		if(!SecRemote.getDepPermission(toAccount))
		{
			System.out.println("Account does not have permission to get balance! ... throwing exception.");
			throw new UnpermittedTransactionException("Get balance Permission denied! exception thrown");
		}
		/** everything seems ok...now try to get the account. */
		Account accountA;
		Account accountB;
		accountA = BankRemote.getAccount(fromAccount.getAccountNum());
		accountB = BankRemote.getAccount(toAccount.getAccountNum());
		/** throw exception if balance is less than money withdrawn. */
		if( amount > accountA.getBalance())
		{
			System.out.println("Account is attempting to withdraw more than it has! ...throwning exception.");
			throw new OverdrawnException("You can withdraw more than you have! Exception thrown!");
		}
		/** synchronize this for multiple user capability. */		
		synchronized(this)
		{
			accountA.withdraw(amount);
			accountB.deposit(amount);
			/** Notify transaction to clients. */
			notifyAll("Transferred $"+amount+" from account " + fromAccount.getAccountNum() +
						" to account " + toAccount.getAccountNum()+".");
		}
	}

}